@@ -14,9 +14,6 @@
 # OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 # PERFORMANCE OF THIS SOFTWARE.
 #
-# $Id: xpmpd.py 589 2010-01-15 05:00:12Z pselkirk $
-#
-
 """
 Extended NAT-PMP daemon
 
@@ -25,6 +22,9 @@ Francis_Dupont@isc.org, November 2009
 Usage: [-d] -m minport -M maxport [-n natsrc]+
 """   
 
+__version__ = '$Id: xpmpd.py 998 2010-11-24 15:06:10Z fdupont $'
+
+
 import getopt
 import heapq
 import select
@@ -62,7 +62,7 @@ class Sockets(object):
 class Mapping(object):
     """Port mapping"""
     __slots__ = ['client', 'src', 'sport', 'nport',
-                 'expire', 'proto', 'companion']
+                 'expire', 'proto', 'companion', 'down']
 
     def __init__(self, client, nport, proto):
         """initializer"""
@@ -73,6 +73,17 @@ class Mapping(object):
         self.expire = 0
         self.proto = proto
         self.companion = True
+        self.down = False
+
+    def hold(self):
+        """hold down a mapping"""
+        self.down = True
+        self.expire = int(time.time()) + 120
+        heapq.heappush(HEAP, self.expire)
+        if EXPIRES.get(self.expire) is None:
+            EXPIRES[self.expire] = [self]
+        else:
+            EXPIRES[self.expire].append(self)
 
 class Client(object):
     """Client"""
@@ -97,15 +108,13 @@ class Client(object):
 
     def getsrc(self, src, sport, proto):
         """search the mapping with src/sport"""
-        candidate = None
         for mapping in self.mappings:
             if (mapping.src != src) or (mapping.proto != proto):
                 continue
             if mapping.companion:
-                candidate = mapping
-            elif (mapping.sport == sport):
+                continue
+            if (mapping.sport == sport):
                 return mapping
-        return candidate
 
     def getcompanion(self, mapping):
         """get the companion of a mapping"""
@@ -117,6 +126,20 @@ class Client(object):
         if candidate.companion:
             return candidate
 
+    def delsrcall(self, src, proto):
+        """delete all mappings for a source"""
+        for mapping in list(self.mappings):
+            if (mapping.src != src) or (mapping.proto != proto):
+                continue
+            if mapping.companion:
+                continue
+            if mapping.down:
+                continue
+            askdelsnat(self, mapping)
+            if mapping.expire != 0:
+                EXPIRES[mapping.expire].remove(mapping)
+            mapping.hold()
+
     def delall(self, proto):
         """delete all mappings"""
         for mapping in list(self.mappings):
@@ -124,16 +147,12 @@ class Client(object):
                 continue
             if mapping.companion:
                 continue
+            if mapping.down:
+                continue
             askdelsnat(self, mapping)
             if mapping.expire != 0:
                 EXPIRES[mapping.expire].remove(mapping)
-            companion = self.getcompanion(mapping)
-            if companion:
-                self.mappings.remove(companion)
-                self.mappings.remove(mapping)
-                NATTED[self.nsrc].ports[mapping.nport] = ''
-            else:
-                mapping.companion = True
+            mapping.hold()
 
 class Natsrc(object):
     """per natted address object, clients per IPv6, IPv6 or '' per port"""
@@ -278,7 +297,7 @@ def delclient(client):
 def getextaddr(ipv6):
     """get external address"""
     client = CLIENTS.get(ipv6)
-    if not client is None:
+    if client is not None:
         return client
     nsrc = asktunnel(ipv6)
     if nsrc != '':
@@ -292,7 +311,8 @@ def expire1(mapping):
     """expire a mapping"""
     client = mapping.client
     debug('expire')
-    askdelsnat(client, mapping)
+    if not mapping.down:
+        askdelsnat(client, mapping)
     companion = client.getcompanion(mapping)
     if companion:
         client.mappings.remove(companion)
@@ -307,32 +327,31 @@ def delmapping(ipv6, src, sport, proto):
     if client is None:
         debug("can't get client in delmapping for %s" % ipv6)
         return None
-    natsrc = NATTED[client.nsrc]
     mapping = client.getsrc(src, sport, proto)
     if mapping is None:
         debug('delmapping: no mapping')
         return client
-    if mapping.companion:
-        debug('delmapping: companion')
+    if mapping.down:
+        debug('delmapping: on hold')
         return client
     askdelsnat(client, mapping)
     if mapping.expire != 0:
         EXPIRES[mapping.expire].remove(mapping)
-    companion = client.getcompanion(mapping)
-    if companion:
-        client.mappings.remove(companion)
-        client.mappings.remove(mapping)
-        natsrc.ports[mapping.nport] = ''
-        debug('delmapping: deleted')
-    else:
-        mapping.companion = True
-        debug('delmapping: kept as companion')
+    mapping.hold()
+    debug('delmapping: hold')
+    return client
+
+def delsrcall(ipv6, src, proto):
+    """delete all mappings for a source"""
+    client = CLIENTS.get(ipv6)
+    if client is not None:
+        client.delsrcall(src, proto)
     return client
 
 def delall(ipv6, proto):
-    """delete all mappings"""
+    """delete all mappings for a client"""
     client = CLIENTS.get(ipv6)
-    if not client is None:
+    if client is not None:
         client.delall(proto)
     return client
 
@@ -344,6 +363,12 @@ def setmapping(ipv6, src, sport, natspec, life, proto):
         return None
     natsrc = NATTED[client.nsrc]
     nport = natspec.get()
+    mapping = client.getsrc(src, sport, proto)
+    if (mapping is not None) and (nport != mapping.nport):
+        if natspec.islocked():
+            raise UnavailablePort
+        natspec.set(mapping.nport)
+        return setmapping(ipv6, src, sport, natspec, life, proto)
     if nport == 0:
         if natspec.islocked():
             raise UnavailablePort
@@ -351,9 +376,9 @@ def setmapping(ipv6, src, sport, natspec, life, proto):
         return setmapping(ipv6, src, sport, natspec, life, proto)
     holder = None
     hentry = natsrc.ports.get(nport)
-    if (not hentry is None) and (hentry != ''):
+    if (hentry is not None) and (hentry != ''):
         holder = natsrc.clients[hentry]
-    if (hentry is None) or ((not holder is None) and (holder != client)):
+    if (hentry is None) or ((holder is not None) and (holder != client)):
         if natspec.islocked():
             raise UnavailablePort
         natspec.set(natsrc.getfreeport())
@@ -410,7 +435,10 @@ def setmapping(ipv6, src, sport, natspec, life, proto):
             debug('setmapping: failed (companion)')
             return None
     # else not mapping.companion
-    if mapping.sport != sport:
+    if mapping.down:
+        mapping.down = False
+        mapping.sport = sport
+    elif mapping.sport != sport:
         if natspec.islocked():
             raise UnavailablePort
         natspec.set(natsrc.getfreeport())
@@ -442,7 +470,7 @@ def senderr(error, sockaddr, packet):
     response += struct.pack('!H', error)
     if len(packet) > 4:
         client = CLIENTS.get(sockaddr[0])
-        if not client is None:
+        if client is not None:
             response += struct.pack('!I', client.sssoe())
             response += packet[4:]
     debug('senderr')
@@ -477,21 +505,38 @@ def getrequest(sockaddr, request):
         response += socket.inet_aton(client.nsrc)
         sendresp(sockaddr, response)
     elif (opcode == 1) or (opcode == 2) or (opcode == 3) or (opcode == 4):
-        locked = (opcode == 3) or (opcode == 4)
         if len(packet) != 12:
             debug('bad set mapping')
             return
         packet = packet[4:]
         (sport, nport, life) = struct.unpack('!HHI', packet)
+        if opcode == 1:
+            proto = 1
+            locked = False
+        elif opcode == 2:
+            proto = 2
+            locked = False
+        elif opcode == 3:
+            proto = 1
+            locked = True
+        else:
+            proto = 2
+            locked = True
         if life == 0:
             nport = 0
             if sport == 0:
-                client = delall(ipv6, opcode)
-                if client is None:
-                    debug("can't find client for delall: %s" % ipv6)
-                    return
+                if src == '0.0.0.0':
+                    client = delall(ipv6, proto)
+                    if client is None:
+                        debug("can't find client for delall: %s" % ipv6)
+                        return
+                else:
+                    client = delsrcall(ipv6, src, proto)
+                    if client is None:
+                        debug("can't find client for delsrcall: %s" % ipv6)
+                        return
             else:
-                client = delmapping(ipv6, src, sport, opcode)
+                client = delmapping(ipv6, src, sport, proto)
                 if client is None:
                     debug("can't find client for delmapping: %s" % ipv6)
                     return
@@ -506,7 +551,7 @@ def getrequest(sockaddr, request):
             if life > 86400:
                 life = 86400
             try:
-                client = setmapping(ipv6, src, sport, natspec, life, opcode)
+                client = setmapping(ipv6, src, sport, natspec, life, proto)
             except OutofResources:
                 senderr(4, sockaddr, request)
                 return
@@ -533,7 +578,7 @@ def expire():
         first = HEAP[0]
         if first > now:
             return
-        while first == HEAP[0]:
+        while (len(HEAP) > 0) and (first == HEAP[0]):
             heapq.heappop(HEAP)
         for mapping in EXPIRES[first]:
             expire1(mapping)
